-----------------------------------------------------------------------------------------------
-- Client Lua Script for BreadCrumbs, by lyravega
-- Wildstar Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
-- Optionless features (set to false in order to disable them, will make entries in the options menu later on): 
local combatUpdates_glocal = true 		--whenever someone receives damage or healing to health, shields, or an absorb shield is applied to them, their bars and stats will be updated 
local showTarget_glocal = true 			--whenever you choose a target, his/her bread will be shown. still subjected to "ignore" filters, will have its own filters later on
local smoothBarUpdates_glocal = true	--whenever any bar has to change, it'll be done smoothly
local lowWarningPulse_glocal = true	--if shieldMode is 0 or 2 (off or advanced), falling below lowHealthThreshold or 30% shields will show a red pulse animation on those bars
-- Delete these (me, not you) when their options are ready, and uncomment them in and around Glocalizer

--TODO List: D-Done, H-Halfway, T-Needs Testing, N-Not Yet Started, S-Scrapped, c-Changed, r-Requires tidy-up
--(H) clickable breads (lacks: expansion later on, customizable buttons & combinations)
--(H) combat events to help soft updates (lacks: option to adjust logic)
--(N) option to track dps/healer/tanks separately
--(H) interrupt armor (lacks: logic)
--(H) target indicator & tracking (lacks: logic)
--(H) custom focus scale (lacks: logic not linked)
--(cD) option for absorb shield location (change: on shieldsMode == 2 and 0, absorb will be shown on health, and will use health/shield max values as its max)
--(cN) multiple focus targets (change: scrapped the idea, instead a whitelist and a blacklist will be introduced)
--(H) pin icons (change: group lacks pin, later need to create my own pins and get rid of reliance on vanilla pin-bar)
--(rD) glocalization (need to compress code and make it look tidy)
--(S) tracking for weapon (reason: for now, maybe later I'll add it)
--(cD) filter type; only health/shield or both (change: shieldsMode == 2 will also track shields in addition to health. may need to separate it in the future)
--(cD) separate override and filter settings for all bread types (change: override is scrapped, too much work and not reliable. just makes a gigantic overhead in the functions)

--Absorb_SP can be used for shield flashes below 30% on shieldsMode == 2

require "Window"
require "GameLib"
--require "Vector2"
require "Vector3"
 
-----------------------------------------------------------------------------------------------
-- BreadCrumbs Module Definition
-----------------------------------------------------------------------------------------------
local BreadCrumbs = {} --do not forget to set this to "local" after testing
local debugMode, debugLevel = true, 1

-----------------------------------------------------------------------------------------------
-- Local (Helper) Functions
-----------------------------------------------------------------------------------------------
local function t_copy(source, target, shallow)
	shallow = target == nil and false or shallow or false
	target = target or {}
    
	if shallow then for key, value in next, source do target[key] = value end
	else for key, value in next, source do
		if type(value) == "table" and type(target[key]) == "table" then t_copy(source[key], target[key])
		elseif type(value) == "table" then target[key] = t_copy(source[key])
		else target[key] = value end
	end end
	
	return target
end

local function t_update(target, source, shallow)
	if shallow then for key, value in next, target do target[key] = source[key] or value end
	else for key, value in next, target do
		if type(value) == "table" and type(source[key]) == "table" then t_update(target[key], source[key])
		elseif source[key] == nil then target[key] = value
		else target[key] = source[key] end
	end end
end

local function t_concat_keys(source) --returns a string of concatenated keys instead of values
	local output = ""
	
	for key in next, source do output = output..key end
	
	return output
end

local function t_size(source, shallow) --number of table elements
	local size = 0
	
	if shallow then for key, value in next, source do size = size + 1 end
	else for key, value in next, source do 
		if type(source[key]) == "table" then size = size + t_size(source[key])
		else size = size + 1 end
	end end
	
	return size
end
 
-----------------------------------------------------------------------------------------------
-- "Glocalization" & Upvalues // Why? Because fuck metatables.
-----------------------------------------------------------------------------------------------
local unitsTable_glocal
local unitsMasterTable_glocal
local breadsTable_glocal
local optionsTable_glocal
local controlsTable_glocal
local internalTable_glocal

local detailedPins_glocal

local distanceCoefficient_glocal
local distanceThreshold_glocal
local distanceFalloff_glocal
local falloffDistance_glocal
local falloffOpacity_glocal

local scale_glocal
local opacity_glocal

local lowHealthThreshold_glocal
local averageHealthThreshold_glocal

local relationshipIcons_glocal

local clickableFrames_glocal
local colorfulClassIcons_glocal
local shieldsMode_glocal
local hybridMode_glocal

local crumbs_glocal
local crumbInstantTransition_glocal
local crumbAmount_glocal
local crumbTick_glocal

local nextFrame_glocal

-- local combatUpdates_glocal
-- local showTarget_glocal
-- local smoothBarUpdates_glocal
-- local lowWarningPulse_glocal

local hostile_glocal
local ally_glocal
local focus_glocal
local group_glocal

local processVector3_Ally
local processVector3_Hostile
local processVector3_Focus
local processVector3_Group

function BreadCrumbs:_function_Glocalizer()
	unitsTable_glocal =				self._tableUnits
	unitsMasterTable_glocal =			self._tableUnitsMaster
	breadsTable_glocal =				self._tableBreads
	optionsTable_glocal = 			self._tableOptions
	controlsTable_glocal = 			self._tableControls
	internalTable_glocal = 			self._tableInternal

	detailedPins_glocal = 				optionsTable_glocal.generalOptions.detailedPins

	distanceCoefficient_glocal = 		optionsTable_glocal.generalOptions.distanceCoefficient
	distanceThreshold_glocal = 		optionsTable_glocal.generalOptions.distanceThreshold
	distanceFalloff_glocal = 			optionsTable_glocal.generalOptions.distanceFalloff
	falloffDistance_glocal = 			optionsTable_glocal.generalOptions.falloffDistance
	falloffOpacity_glocal = 			optionsTable_glocal.generalOptions.falloffOpacity

	scale_glocal = 					optionsTable_glocal.generalOptions.scale
	opacity_glocal = 				optionsTable_glocal.generalOptions.opacity
	
	lowHealthThreshold_glocal = 		optionsTable_glocal.generalOptions.lowHealthThreshold
	averageHealthThreshold_glocal = 	optionsTable_glocal.generalOptions.averageHealthThreshold
	
	relationshipIcons_glocal = 			optionsTable_glocal.generalOptions.relationshipIcons
	
	clickableFrames_glocal = 			optionsTable_glocal.generalOptions.clickableFrames
	colorfulClassIcons_glocal = 		optionsTable_glocal.generalOptions.colorfulClassIcons
	shieldsMode_glocal = 			optionsTable_glocal.generalOptions.shieldsMode
	hybridMode_glocal = 				optionsTable_glocal.generalOptions.hybridMode
	
	crumbs_glocal = 				optionsTable_glocal.generalOptions.crumbs
	crumbInstantTransition_glocal = 		optionsTable_glocal.generalOptions.crumbInstantTransition
	crumbAmount_glocal = 			optionsTable_glocal.generalOptions.crumbAmount
	crumbTick_glocal = 				optionsTable_glocal.generalOptions.crumbTick
	
	nextFrame_glocal = 				optionsTable_glocal.generalOptions.drawOnNextFrame
	
	-- combatUpdates_glocal = 			optionsTable_glocal.generalOptions.combatUpdates
	-- showTarget_glocal = 				optionsTable_glocal.generalOptions.showTarget
	-- smoothBarUpdates_glocal = 		optionsTable_glocal.generalOptions.smoothBarUpdates
	-- lowWarningPulse_glocal = 		optionsTable_glocal.generalOptions.lowWarningPulse
	
	hostile_glocal = 				optionsTable_glocal.breadOptions.hostile.enabled
	ally_glocal = 					optionsTable_glocal.breadOptions.ally.enabled
	focus_glocal = 					optionsTable_glocal.breadOptions.focus.enabled
	group_glocal = 					optionsTable_glocal.breadOptions.group.enabled
	
	processVector3_Hostile = 			self:_functionBuilder_ProcessVector3(optionsTable_glocal.breadOptions.hostile)
	processVector3_Ally = 			self:_functionBuilder_ProcessVector3(optionsTable_glocal.breadOptions.ally)
	processVector3_Focus = 			self:_functionBuilder_ProcessVector3(optionsTable_glocal.breadOptions.focus)
	processVector3_Group = 			self:_functionBuilder_ProcessVector3(optionsTable_glocal.breadOptions.group)
end

local playerUnit_glocal

local next, pairs, ipairs = next, pairs, ipairs
local t_insert, t_sort, t_remove, t_unpack = table.insert, table.sort, table.remove, unpack --t_insert is not used; instead of table.insert(table, data), table[#table+1]=data is used // switched to "count = 0; count = count + 1; table[count]=data" for some
local s_format, s_gsub = string.format, string.gsub
local m_floor, m_log = math.floor, math.log
local type = type
local Apollo, XmlDoc = Apollo, XmlDoc

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local classIcons = {
	[GameLib.CodeEnumClass.Engineer] = 	"BreadCrumbsSprites:Engineer",
	[GameLib.CodeEnumClass.Esper] = 		"BreadCrumbsSprites:Esper",
	[GameLib.CodeEnumClass.Medic] = 		"BreadCrumbsSprites:Medic",
	[GameLib.CodeEnumClass.Spellslinger] = 	"BreadCrumbsSprites:Spellslinger",
	[GameLib.CodeEnumClass.Stalker] = 		"BreadCrumbsSprites:Stalker",
	[GameLib.CodeEnumClass.Warrior] = 		"BreadCrumbsSprites:Warrior",
} --Colored variants: ..Colored

local relationshipIcons = {
	guild = 							"IconSprites:Icon_MapNode_Map_GroupMember",
	rival = 							"IconSprites:Icon_MapNode_Map_Rival",
	friend = 							"IconSprites:Icon_MapNode_Map_Friend",
}

local relationshipPin = {
	guild = 							"+ ",
	rival = 							"x ",
	friend = 							"o ",
	none = 							"~ ",
}

local dispositionColorTable = {
	focus = 							ApolloColor.new("DispositionNeutral"),
	hostile = 							ApolloColor.new("DispositionHostile"),
	ally = 							ApolloColor.new("DispositionFriendly"),
	group = 							ApolloColor.new("DispositionFriendlyUnflagged"),
}

local barSprites = {
	low = 							"HUD_TargetFrame:spr_TargetFrame_ClusterHealthRed",
	medium = 						"HUD_TargetFrame:spr_TargetFrame_ClusterHealthYellow",
	high = 							"HUD_TargetFrame:spr_TargetFrame_ClusterHealthGreen", --ffa0ffd0
	shield = 							"HUD_TargetFrame:spr_TargetFrame_ClusterHealthShield",
	shieldLow = 						"BreadCrumbsSprites:ShieldLowPulse",
}

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function BreadCrumbs:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self 
	
	o._tableControls = {
		generalControls = {},
		breadControls = {
			hostile = {},
			ally = {},
			group = {},
			focus = {},
		},
	}
	o._tableBreads = {}
--	o._tableCrumbs = {}
	o._tablePlayer = {}
	o._tableUnitsMaster = {}
	o._tableUnits = {
		hostile = {}, 
		ally = {}, 
		group = {}, 
		focus = {}, 
--		weapon = {}, 
--		specials = {},
	}
	o._tableOptions = {
		generalOptions = {
			detailedPins = true,
			distanceCoefficient = 1.5,
			distanceThreshold = 200,
			distanceFalloff = true,
			falloffDistance = 50,
			falloffOpacity = 0.5,
			scale = 1,
			opacity = 0.85,
			lowHealthThreshold = 0.3,
			averageHealthThreshold = 0.7,
			relationshipIcons = true,
			colorfulClassIcons = false, 
			shieldsMode = 1, 
			hybridMode = 2, 
			clickableFrames = false, 
			crumbs = true,
			crumbInstantTransition = false,
			crumbAmount = 2,
			crumbTick = 15,
			drawOnNextFrame = true,
			primaryTimer = 100, --Hz; 100 = 0.01
			secondaryTimer = 8, --Hz; 4 = 0.25
			combatUpdates = true, --lacks options
			showTarget = true, --lacks options
			smoothBarUpdates = true, --lacks options
			lowWarningPulse = true, --lacks options
		},
		breadOptions = {
			hostile = {
				breadType = "hostile",
				enabled = true,
				maximumAmount = 1,
				ignoreOffscreen = false,
				ignoreOccluded = false,
				ignoreDead = true,
				ignoreAbove = false,
				ignoreThreshold = 0.9,
			},
			ally = {
				breadType = "ally",
				enabled = true,
				maximumAmount = 1,
				ignoreOffscreen = false,
				ignoreOccluded = false,
				ignoreDead = false,
				ignoreAbove = false,
				ignoreThreshold = 0.9,
			},
			group = {
				breadType = "group",
				enabled = true,
				maximumAmount = 3,
				ignoreOffscreen = false,
				ignoreOccluded = false,
				ignoreDead = false,
				ignoreAbove = false,
				ignoreThreshold = 0.9,
			},
			focus = {
				breadType = "focus",
				enabled = true,
				maximumAmount = 1,
				ignoreOffscreen = false,
				ignoreOccluded = false,
				ignoreDead = false,
				ignoreAbove = false,
				ignoreThreshold = 0.9,
			},
		},
		-- basicOptions = {
		-- },
		-- advancedOptions = {
		-- },
		-- expertOptions = {
		-- },
		menuOffsets = {369,68,863,744},
		restoreDefaults = false,
		configVersion = 1,
	}
	o._tableInternal = {
		initializeTimer = 0.500,
		addOnInitialized = false,
		processLater = {},
		pinDetail = {
			hostile = {},
			ally = {}, 
			group = {}, 
		},
		nameLists = {
			blacklist = {},
			whitelist = {},		
		},
	}

	return o
end

-----------------------------------------------------------------------------------------------
-- BreadCrumbs Init & Load
-----------------------------------------------------------------------------------------------
function BreadCrumbs:Init()
	Apollo.RegisterAddon(self, true, "BreadCrumbs")
	
	Apollo.RegisterSlashCommand("BreadCrumbs", 		"_windowFunction_ProcessMenu", self)
	Apollo.RegisterSlashCommand("bc", 				"_windowFunction_ProcessMenu", self)
end

function BreadCrumbs:OnLoad()
	Apollo.LoadSprites("BreadCrumbsSprites.xml")
	
	self.xmlDocument = XmlDoc.CreateFromFile("BreadCrumbs.xml")
	self.xmlDocument:RegisterCallback("OnDocLoaded", self)
	
	self._initializeTimer = ApolloTimer.Create(self._tableInternal.initializeTimer, true, "_handler_Initialize", self)
	
	Apollo.RegisterEventHandler("ChangeWorld", 		"_handler_ChangeWorld", self)
	Apollo.RegisterEventHandler("UnitCreated", 			"_handler_UnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", 		"_handler_UnitDestroyed", self)
--	Apollo.RegisterEventHandler("Group_Updated", 		"_handler_GroupUpdated", self)
--	Apollo.RegisterEventHandler("CharacterCreated",		"_handler_CharacterCreated", self)
	
	if Apollo.GetAddon("InterfaceMenuList") then --dead man's switch; if it is disabled or replaced, don't try to update shit as it is not necessary
		Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", 		"_handler_InterfaceMenuListHasLoaded", self)
		Apollo.RegisterEventHandler("_event_ToggleAllies", 			"_handler_ToggleAllies", self) --user-made event
		Apollo.RegisterEventHandler("_event_ToggleHostiles", 			"_handler_ToggleHostiles", self) --user-made event
		Apollo.RegisterEventHandler("_event_ToggleGroup", 			"_handler_ToggleGroup", self) --user-made event
	end 
	
	if debugMode and Apollo.GetAddon("Rover") then SendVarToRover("BreadCrumbs", self) end
end

function BreadCrumbs:OnDelayedLoad() --function load order is: OnLoad; GetAsyncLoadStatus; OnRestore; OnDelayedLoad (OnDelayedLoad is user-made)
	self:_function_Glocalizer()
	internalTable_glocal.addOnInitialized = true
	
	if focus_glocal then 			Apollo.RegisterEventHandler("AlternateTargetUnitChanged", 		"_handler_FocusChanged", self) end
	if combatUpdates_glocal then 	Apollo.RegisterEventHandler("DamageOrHealingDone", 		"_handler_CombatAction", self) end
	if showTarget_glocal then 		Apollo.RegisterEventHandler("TargetUnitChanged", 			"_handler_TargetChanged", self) end
	if nextFrame_glocal then 		Apollo.RegisterEventHandler("NextFrame", 					"_handler_MainProcess", self)
	else self._primaryTimer = 		ApolloTimer.Create(1/optionsTable_glocal.generalOptions.primaryTimer, true, "_handler_MainProcess", self) end
	self._secondaryTimer = 		ApolloTimer.Create(1/optionsTable_glocal.generalOptions.secondaryTimer, true, "_handler_AuxProcess", self)
end

function BreadCrumbs:OnDocLoaded()
	if self.xmlDocument ~= nil and self.xmlDocument:IsLoaded() then
		self.optionsWindow = Apollo.LoadForm(self.xmlDocument, "BreadCrumbs_Options", nil, self)
		
		if self.optionsWindow == nil then Apollo.AddAddonErrorText(self, "Could not load the main window for some reason."); return end
		self.optionsWindow:Show(false, true)
	end
	
	-- local test = Apollo.LoadForm(self.xmlDocument, "Test", "InWorldHudStratum", self)
	-- self.testBar = test:FindChild("Cast")
	-- self.testBar:SetMax(100)
end

-----------------------------------------------------------------------------------------------
-- BreadCrumbs Save & Restore
-----------------------------------------------------------------------------------------------
function BreadCrumbs:OnSave(saveDepth)
	if saveDepth ~= GameLib.CodeEnumAddonSaveLevel.Character then return nil end

	local savedVariables = t_copy(self._tableOptions)

	return savedVariables
end

function BreadCrumbs:OnRestore(saveDepth, savedVariables) --OnLoad then GetAsyncLoad then OnRestore
	if saveDepth ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	
	if savedVariables and not savedVariables.restoreDefaults and savedVariables.configVersion == self._tableOptions.configVersion then t_update(self._tableOptions, savedVariables) end; self:_function_Glocalizer()
end

-----------------------------------------------------------------------------------------------
-- BreadCrumbs Event Handlers
-----------------------------------------------------------------------------------------------
function BreadCrumbs:_handler_Initialize()
	self._tablePlayer.unit = GameLib.GetPlayerUnit(); if not self._tablePlayer.unit then return end
	playerUnit_glocal = self._tablePlayer.unit

	if not self._tableInternal.addOnInitialized then self:OnDelayedLoad() end
	
	self._initializeTimer:Stop()
	if not nextFrame_glocal then self._primaryTimer:Start() end
	self._secondaryTimer:Start()
	
	if next(internalTable_glocal.processLater) then self:_function_ProcessLater() end
end

function BreadCrumbs:_handler_ChangeWorld() --damn you playerunit. Even carbine has to hack in order to get you properly...
	self._tablePlayer.unit = nil
	
	if not nextFrame_glocal then self._primaryTimer:Stop() end
	self._secondaryTimer:Stop()
	self._initializeTimer:Start() --reinitialize, actually. 
end

function BreadCrumbs:_handler_MainProcess()
	local pvector3 = Vector3.New(playerUnit_glocal:GetPosition())
	
	if hostile_glocal then processVector3_Hostile(pvector3) end
	if ally_glocal then processVector3_Ally(pvector3) end
	if focus_glocal then processVector3_Focus(pvector3) end
	if group_glocal then processVector3_Group(pvector3) end
end

function BreadCrumbs:_handler_AuxProcess()
	self:_function_ProcessWindow()
end

function BreadCrumbs:_handler_UnitCreated(newUnit)
	if not newUnit then return end
	if not newUnit:IsACharacter() then return end -- no weapon support for now. for weapon: newUnit:GetType() == "Pickup" and s_find(newUnit:GetName(), self._tablePlayer.name)
	if not self._tablePlayer.unit then internalTable_glocal.processLater[#internalTable_glocal.processLater+1]=newUnit; return end --table.insert
	
	local newUnitId = newUnit:GetId(); if unitsMasterTable_glocal[newUnitId] then return end
	local newUnitName = newUnit:GetName()
	local breadType
	
	if newUnit:IsThePlayer() then return
	elseif focus_glocal and newUnit == self._tablePlayer.unit:GetAlternateTarget() then breadType = "focus"
	elseif group_glocal and newUnit:IsInYourGroup() then breadType = "group"
	elseif newUnit:GetFaction() == self._tablePlayer.unit:GetFaction() then breadType = "ally"
	else breadType = "hostile" end
	
	-- if newUnit:IsThePlayer() or blacklist_glocal[newUnitName] then return
	-- elseif whitelist_glocal[newUnitName] then breadType = "greenlit"
	-- elseif focus_glocal and newUnit == self._tablePlayer.unit:GetAlternateTarget() then breadType = "focus"
	-- elseif group_glocal and newUnit:IsInYourGroup() then breadType = "group"
	-- elseif newUnit:GetFaction() == self._tablePlayer.unit:GetFaction() then breadType = "ally"
	-- else breadType = "hostile" end
	
	local color = dispositionColorTable[breadType]
	
	unitsTable_glocal[breadType][newUnitId] = {
		unit = 		newUnit, 
		unitId = 		newUnitId,
		name = 		newUnitName, 
		isOnScreen = 	GameLib.GetUnitScreenPosition(newUnit).bOnScreen,
		isOccluded = 	newUnit:IsOccluded(),
		isTarget = 	false,
		-- class = 		newUnit:GetClassId(),
		relationship = 	(newUnit:IsFriend() or newUnit:IsAccountFriend()) and "friend" or newUnit:IsRival() and "rival" or newUnit:GetGuildName() and newUnit:GetGuildName() == self._tablePlayer.unit:GetGuildName() and "guild" or "none",
		vitals = {
			percHP = 	newUnit:GetHealth() / newUnit:GetMaxHealth(),
			percSP = 	newUnit:GetShieldCapacity() / newUnit:GetShieldCapacityMax(),
		},
	}; local newUnitTable = unitsTable_glocal[breadType][newUnitId]
	
	newUnitTable.crumbs = {}
	if crumbs_glocal then for crumbNumber = 1, crumbAmount_glocal do
		local crumb = Apollo.LoadForm(self.xmlDocument, "Crumb", "InWorldHudStratum", self); newUnitTable.crumbs[crumbNumber] = crumb
		crumb:SetBGColor(color)
		
		crumb:SetScale(scale_glocal)
		crumb:SetOpacity(opacity_glocal)
		crumb:Show(false) 
	end end
	
	local bread = Apollo.LoadForm(self.xmlDocument, "Bread", "InWorldHudStratum", self); newUnitTable.bread = bread
	newUnitTable.breadCrumbs = {
		HP = 		bread:FindChild("Health"),
		HPulse = 		bread:FindChild("Health_Pulse"),
		SP = 		bread:FindChild("Shield"),
		SPulse = 		bread:FindChild("Shield_Pulse"),
		AP = 		bread:FindChild("Absorb"),
		Name = 		bread:FindChild("Name"),
		Distance = 	bread:FindChild("Distance"),
		Class = 		bread:FindChild("Class"),
		_BG = 		bread:FindChild("BG"),
		_FG = 		bread:FindChild("Cover"),
		Relationship = 	bread:FindChild("Relationship"),
		Target = 		bread:FindChild("Target"),
		Overload = 	bread:FindChild("Overload"),
		Button = 		bread:FindChild("Button"),
	}; local breadCrumbs = newUnitTable.breadCrumbs
	
	bread:SetScale(scale_glocal)
	bread:SetOpacity(opacity_glocal)
	bread:Show(false)
	
	breadCrumbs.HPulse:SetMax(1); breadCrumbs.HPulse:SetProgress(1)
	breadCrumbs.SPulse:SetMax(1); breadCrumbs.SPulse:SetProgress(1)
	breadCrumbs._BG:SetBGColor(color)
	breadCrumbs._FG:SetBGColor(color)
	breadCrumbs.Target:SetBGColor(color)
	breadCrumbs.Name:SetText(newUnitTable.name); breadCrumbs.Name:SetTextColor(color)
	if clickableFrames_glocal then breadCrumbs.Button:Show(true); newUnitTable.bread:SetData(newUnit) end
	if colorfulClassIcons_glocal then breadCrumbs.Class:SetSprite(classIcons[newUnit:GetClassId()].."_Colored") else breadCrumbs.Class:SetSprite(classIcons[newUnit:GetClassId()]); breadCrumbs.Class:SetBGColor(color) end
	if relationshipIcons_glocal then breadCrumbs.Relationship:SetSprite(relationshipIcons[newUnitTable.relationship]); breadCrumbs.Relationship:Show(true) end
	if shieldsMode_glocal == 0 then 
		breadCrumbs._FG:SetSprite("HUD_TargetFrame:spr_TargetFrame_ClusterCover") 
		breadCrumbs.HP:SetRadialMax(310)
		breadCrumbs.HPulse:SetRadialMax(310)
		breadCrumbs.AP:SetRadialMax(310)
		breadCrumbs.SP:Show(false)
		breadCrumbs.SPulse:Show(false) --already hidden, but meh.
	end
	
	-- if detailedPins_glocal and (breadType == "ally" or breadType == "hostile" or breadType == "group") then
	if detailedPins_glocal and breadType ~= "focus" then
		local pinName = relationshipPin[newUnitTable.relationship]..newUnitName.."\n"
		if breadType == "ally" then internalTable_glocal.pinDetail.ally[pinName] = true; self:_function_UpdateAllyPin()
		elseif breadType == "hostile" then internalTable_glocal.pinDetail.hostile[pinName] = true; self:_function_UpdateHostilePin() 
		elseif breadType == "group" then internalTable_glocal.pinDetail.group[pinName] = true; self:_function_UpdateGroupPin() end
	end

	unitsMasterTable_glocal[newUnitId] = breadType
end

function BreadCrumbs:_handler_UnitDestroyed(oldUnit)
	if not oldUnit then return end
	
	local oldUnitId = oldUnit:GetId(); if not unitsMasterTable_glocal[oldUnitId] then return end
	local breadType = unitsMasterTable_glocal[oldUnitId]
	local oldUnitTable = unitsTable_glocal[breadType][oldUnitId]
	
	oldUnitTable.bread:Destroy() --can make a table to avoid destroying and instead reusing stuff
	if crumbs_glocal then for crumbNumber = 1, crumbAmount_glocal do
		oldUnitTable.crumbs[crumbNumber]:Destroy()
	end end
	
	-- if detailedPins_glocal and (breadType == "ally" or breadType == "hostile" or breadType == "group") then
	if detailedPins_glocal and breadType ~= "focus" then
		local pinName = relationshipPin[oldUnitTable.relationship]..oldUnitTable.name.."\n"
		if breadType == "ally" then internalTable_glocal.pinDetail.ally[pinName] = nil; self:_function_UpdateAllyPin()
		elseif breadType == "hostile" then internalTable_glocal.pinDetail.hostile[pinName] = nil; self:_function_UpdateHostilePin()
		elseif breadType == "group" then internalTable_glocal.pinDetail.group[pinName] = nil; self:_function_UpdateGroupPin() end
	end
	
	unitsTable_glocal[breadType][oldUnitId] = nil --stores unitTables
	unitsMasterTable_glocal[oldUnitId] = nil --stores unitIds and their breadTypes
	breadsTable_glocal[oldUnitId] = nil --stores show/hide data
end

function BreadCrumbs:_handler_TargetChanged(targetUnit) --needs more work; needs option to enable, and need a rewrite to avoid using the for loop, and instead just making isTarget false for the old target only
	for unitId in next, breadsTable_glocal do
		local unitTable = unitsTable_glocal[unitsMasterTable_glocal[unitId]][unitId]
		
		if unitTable.isTarget then unitTable.isTarget = false
		unitTable.breadCrumbs.Target:Show(false) end
	end
	
	local targetUnitId; if targetUnit then targetUnitId = targetUnit:GetId() end
	
	if unitsMasterTable_glocal[targetUnitId] then 
		local targetUnitTable = unitsTable_glocal[unitsMasterTable_glocal[targetUnitId]][targetUnitId]
		
		targetUnitTable.isTarget = true
		targetUnitTable.breadCrumbs.Target:Show(true)
	end
end

function BreadCrumbs:_handler_FocusChanged(focusUnit)
	local _, oldFocus, oldFocusUnit; if next(unitsTable_glocal.focus) then _, oldFocus = next(unitsTable_glocal.focus); oldFocusUnit = oldFocus.unit end

	if oldFocusUnit then self:_handler_UnitDestroyed(oldFocusUnit); self:_handler_UnitCreated(oldFocusUnit) end
	if focusUnit then self:_handler_UnitDestroyed(focusUnit); self:_handler_UnitCreated(focusUnit) end
end

--GroupLib.GetUnitForGroupMember(index)
--GroupLib.GetMemberCount()

-- function BreadCrumbs:_handler_GroupUpdated() --need testing urgently
	-- for unitId, breadType in next, unitsMasterTable_glocal do
		-- if breadType == "group" or breadType == "ally" then
			-- local unitTable = unitsTable_glocal[breadType][unitId]
			
			-- if breadType == "group" and not unitTable.unit:IsInYourGroup() then self:_handler_UnitDestroyed(unitTable.unit); self:_handler_UnitCreated(unitTable.unit) --deserter
			-- elseif breadType == "ally" and unitTable.unit:IsInYourGroup() then self:_handler_UnitDestroyed(unitTable.unit); self:_handler_UnitCreated(unitTable.unit) end --new arrival
		-- end
	-- end
-- end

--	DamageOrHealingDone(unitCaster,unitTarget,eDamageType,nHealth,nShields,nAbsorb,bIsCritical,splUsed) - fires whenever a unit takes damage or gets healed 
--	if casterUnit and unitsMasterTable_glocal[casterUnit:GetId()] then self:_function_ProcessUnit(casterUnit) end
function BreadCrumbs:_handler_CombatAction(_, targetUnit) --doesn't track interrupt armor or shit as expected
	if not targetUnit then return end
	
	local targetUnitId = targetUnit:GetId(); if not breadsTable_glocal[targetUnitId] then return end
	
	self:_function_ProcessUnit(targetUnit, unitsTable_glocal[unitsMasterTable_glocal[targetUnitId]][targetUnitId], breadsTable_glocal[targetUnitId])
end

-----------------------------------------------------------------------------------------------
-- BreadCrumbs Pin Event Handlers and Functions
-----------------------------------------------------------------------------------------------
function BreadCrumbs:_handler_InterfaceMenuListHasLoaded() -- NTS: can malfunction; add-on waits for interface to load; if interface loads before, what then?
	if not self._tableInternal.addOnInitialized then return end -- NTS: dead man's switch to avoid the malfunction above, better not running it than running it along with errors

	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "BreadCrumbs Allies", {"_event_ToggleAllies", "", relationshipIcons.friend}); self:_function_UpdateAllyPin()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "BreadCrumbs Hostiles", {"_event_ToggleHostiles", "", relationshipIcons.rival}); self:_function_UpdateHostilePin() 
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "BreadCrumbs Group", {"_event_ToggleGroup", "", relationshipIcons.guild}); self:_function_UpdateGroupPin() 
end

function BreadCrumbs:_handler_ToggleAllies()
	ally_glocal = not ally_glocal
	
	if not ally_glocal then self:_function_PinCleanUp(unitsTable_glocal.ally) end
	
	local temp = ally_glocal ~= optionsTable_glocal.breadOptions.ally.enabled and " Temporarily " or " "
	local message = not ally_glocal and "Ally Tracking"..temp.."Disabled" or "Ally Tracking"..temp.."Enabled"
	ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Datachron, message, "BreadCrumbs" )
	self:_function_UpdateAllyPin()
end

function BreadCrumbs:_handler_ToggleHostiles()
	hostile_glocal = not hostile_glocal
	
	if not hostile_glocal then self:_function_PinCleanUp(unitsTable_glocal.hostile) end
	
	local temp = hostile_glocal ~= optionsTable_glocal.breadOptions.hostile.enabled and " Temporarily " or " "
	local message = not hostile_glocal and "Hostile Tracking"..temp.."Disabled" or "Hostile Tracking"..temp.."Enabled"
	ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Datachron, message, "BreadCrumbs" )
	self:_function_UpdateHostilePin()
end

function BreadCrumbs:_handler_ToggleGroup()
	group_glocal = not group_glocal
	
	if not group_glocal then self:_function_PinCleanUp(unitsTable_glocal.group) end
	
	local temp = group_glocal ~= optionsTable_glocal.breadOptions.group.enabled and " Temporarily " or " "
	local message = not group_glocal and "Group Tracking"..temp.."Disabled" or "Group Tracking"..temp.."Enabled"
	ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Datachron, message, "BreadCrumbs" )
	self:_function_UpdateGroupPin()
end

function BreadCrumbs:_function_UpdateAllyPin()
	if detailedPins_glocal then Event_FireGenericEvent("InterfaceMenuList_AlertAddOn", "BreadCrumbs Allies", {ally_glocal, "\n"..t_concat_keys(internalTable_glocal.pinDetail.ally), t_size(internalTable_glocal.pinDetail.ally, true)})
	else Event_FireGenericEvent("InterfaceMenuList_AlertAddOn", "BreadCrumbs Allies", {ally_glocal, "", 0}) end
end

function BreadCrumbs:_function_UpdateHostilePin()
	if detailedPins_glocal then Event_FireGenericEvent("InterfaceMenuList_AlertAddOn", "BreadCrumbs Hostiles", {hostile_glocal, "\n"..t_concat_keys(internalTable_glocal.pinDetail.hostile), t_size(internalTable_glocal.pinDetail.hostile, true)})
	else Event_FireGenericEvent("InterfaceMenuList_AlertAddOn", "BreadCrumbs Hostiles", {hostile_glocal, "", 0}) end
end

function BreadCrumbs:_function_UpdateGroupPin()
	if detailedPins_glocal then Event_FireGenericEvent("InterfaceMenuList_AlertAddOn", "BreadCrumbs Group", {group_glocal, "\n"..t_concat_keys(internalTable_glocal.pinDetail.group), t_size(internalTable_glocal.pinDetail.group, true)})
	else Event_FireGenericEvent("InterfaceMenuList_AlertAddOn", "BreadCrumbs Group", {group_glocal, "", 0}) end
end

function BreadCrumbs:_function_PinCleanUp(breadTable)
	for unitId, unitTable in next, breadTable do
		breadsTable_glocal[unitId] = nil
		
		if crumbs_glocal then for crumbNumber = 1, crumbAmount_glocal do
			unitTable.crumbs[crumbNumber]:Show(false, crumbInstantTransition_glocal)
		end end; unitTable.bread:Show(false)	
	end
end

--can make a pin for hybrid

-----------------------------------------------------------------------------------------------
-- BreadCrumbs Functions
-----------------------------------------------------------------------------------------------
function BreadCrumbs:_function_ProcessLater()
	while next(internalTable_glocal.processLater) do self:_handler_UnitCreated(t_remove(internalTable_glocal.processLater)) end
end

function BreadCrumbs:_functionBuilder_ProcessVector3(breadOptions)
	local ignoreDead, ignoreOccluded, ignoreOffscreen 		= breadOptions.ignoreDead, breadOptions.ignoreOccluded, breadOptions.ignoreOffscreen
	local maximumAmount, ignoreAbove, ignoreThreshold 	= breadOptions.maximumAmount, breadOptions.ignoreAbove, breadOptions.ignoreThreshold
	local breadTable, visibilityTable 						= self._tableUnits[breadOptions.breadType], self._tableBreads
	local crumbs, crumbAmount, crumbTick 				= self._tableOptions.generalOptions.crumbs, self._tableOptions.generalOptions.crumbAmount, self._tableOptions.generalOptions.crumbTick
	local distanceCoefficient, distanceThreshold 			= self._tableOptions.generalOptions.distanceCoefficient, self._tableOptions.generalOptions.distanceThreshold
	
	local newFunction = function(pvector3)
		local breadArray = {}
		local unitCount, maxDistance, firstUnit = 0, 0, 1
		
		for unitId, unitTable in next, breadTable do
			local unit = unitTable.unit:GetVehicle() or unitTable.unit
			local healthPerc = unitTable.vitals.percHP
			
			local tvector3 = Vector3.New(unit:GetPosition()); unitTable.vector3 = tvector3
			local distance = (pvector3 - tvector3):Length(); unitTable.distance = distance		
			
			--isTarget needs its own checks here.
			if unitTable.isTarget then breadArray[0] = unitTable; firstUnit = 0 --; maximumAmount = maximumAmount - 1 
			elseif distance < distanceThreshold
			and (distance < maxDistance or unitCount < maximumAmount) --interesting stuff here about target. similar can be used for whitelist and shit?
			and (not ignoreAbove or ignoreAbove and healthPerc < ignoreThreshold)
			and (not ignoreDead or ignoreDead and healthPerc > 0)
			and (not ignoreOccluded or ignoreOccluded and not unitTable.isOccluded)
			and (not ignoreOffscreen or ignoreOffscreen and unitTable.isOnScreen) then
				if unitCount == 0 or breadArray[unitCount].distance < distance then breadArray[unitCount+1] = unitTable
				else for unitIndex = unitCount, 1, -1 do 
					if unitIndex == 1 and breadArray[1].distance > distance then breadArray[2] = breadArray[1]; breadArray[1] = unitTable
					elseif breadArray[unitIndex].distance > distance then breadArray[unitIndex+1] = breadArray[unitIndex]
					else breadArray[unitIndex+1] = unitTable; break end
				end end
				if unitCount < maximumAmount then unitCount = unitCount + 1 end; maxDistance = breadArray[unitCount].distance
			end
			
			visibilityTable[unitId] = false
		end
		
		for unitIndex = firstUnit, unitCount do
			local tvector3 = breadArray[unitIndex].vector3
			local distance = breadArray[unitIndex].distance
			
			breadArray[unitIndex].bread:SetWorldLocation(pvector3 + (tvector3 - pvector3):Normal()*m_log(distance+1.1)^distanceCoefficient)
			
			if crumbs then for crumbNumber = 1, crumbAmount do
				if crumbTick*crumbNumber < distance + 5 then
					breadArray[unitIndex].crumbs[crumbNumber]:SetWorldLocation(pvector3 + (tvector3 - pvector3):Normal()*m_log((distance+1.1)^(crumbNumber+1))^distanceCoefficient)
				end
			end end
		
			visibilityTable[breadArray[unitIndex].unitId] = true
		end
	end
	
	return newFunction
end

function BreadCrumbs:_function_ProcessWindow() --need to make a functionBuilder specific to breadTypes like above. Then, can imitate isTarget above, and use the "old" extra idea.
	for unitId, toDraw in next, breadsTable_glocal do
		local unitTable = unitsTable_glocal[unitsMasterTable_glocal[unitId]][unitId]
		local unit, bread, distance = unitTable.unit, unitTable.bread, unitTable.distance
		local vehicleOrUnit = unit:GetVehicle() or unit
		local hybrid = false
		
		unitTable.isOnScreen = GameLib.GetUnitScreenPosition(unit).bOnScreen
		local isOccluded = vehicleOrUnit:IsOccluded(); unitTable.isOccluded = isOccluded
		-- unitTable.isDead = unit:IsDead()
		
		if hybridMode_glocal > 0 and not toDraw and not isOccluded and distance < 100 then --may need a distance check
			if hybridMode_glocal == 2 and (unitTable.vitals.percHP < lowHealthThreshold_glocal or shieldsMode_glocal and unitTable.vitals.percSP < 0.3)
			or hybridMode_glocal == 1 then bread:SetUnit(unit, 0); hybrid = true end
		end
		
		self:_function_ProcessUnit(unit, unitTable, toDraw or hybrid)
		
		if toDraw or hybrid then
			local breadCrumbs = unitTable.breadCrumbs
			
			breadCrumbs.Distance:SetText(m_floor(distance))
			if distanceFalloff_glocal then bread:SetOpacity(distance > falloffDistance_glocal and falloffOpacity_glocal or opacity_glocal) end
		end
		
		if crumbs_glocal then for crumbNumber = 1, crumbAmount_glocal do
			local crumb = unitTable.crumbs[crumbNumber]
			
			crumb:Show(not hybrid and toDraw and crumb:IsOnScreen() and crumbTick_glocal*crumbNumber < unitTable.distance, crumbInstantTransition_glocal)
			-- if toDraw and distanceFalloff_glocal then crumb:SetOpacity(distance > falloffDistance_glocal and falloffOpacity_glocal or opacity_glocal) end --do this later. crumbs are irrelevant. resistance is futile.
		end end; bread:Show(toDraw or hybrid)
	end
	
	-- if playerUnit_glocal:IsCasting() then --Print("casting")
		-- local castElapsed, castDuration = playerUnit_glocal:GetCastElapsed(), playerUnit_glocal:GetCastDuration()
		-- Print(castElapsed); Print(castDuration)
		-- self.testBar:SetProgress(100, 100000/(castDuration-castElapsed))
	-- else self.testBar:SetProgress(0); --Print("idle")
	-- end
end

function BreadCrumbs:_function_ProcessUnit(unit, unitTable, toDraw) --TODO: learn more about DamageOrHealingDone; nHealth/nShields/nAbsorb doesn't work properly, only nHealth has a value. As said, learn more about it, and update hp/sp/ap accordingly, not all at once.
	local unitVitals = unitTable.vitals
	
	local healthMax = unit:GetMaxHealth()--; unitVitals.maxHP = healthMax
	local healthCurr = unit:GetHealth()--; unitVitals.currHP = healthCurr
	local healthPerc = healthCurr / healthMax; unitVitals.percHP = healthPerc
	local shieldMax = unit:GetShieldCapacityMax()--; unitVitals.maxSP = shieldMax
	local shieldCurr = unit:GetShieldCapacity()--; unitVitals.currSP = shieldCurr
	local shieldPerc = shieldCurr / shieldMax; unitVitals.percSP = shieldPerc
	-- local absorbMax = unit:GetAbsorptionMax(); unitVitals.maxAP = absorbMax
	local absorbMax = healthMax
	local absorbCurr = unit:GetAbsorptionValue()--; unitVitals.currAP = absorbCurr
	
	if toDraw then
		local breadCrumbs = unitTable.breadCrumbs
		local isAlive = healthPerc > 0
		
		breadCrumbs.HP:SetMax(healthMax)
		breadCrumbs.HP:SetProgress(healthCurr, smoothBarUpdates_glocal and healthMax*2 or nil)
		breadCrumbs.HP:SetFullSprite(healthPerc > averageHealthThreshold_glocal and barSprites.high or healthPerc > lowHealthThreshold_glocal and barSprites.medium or barSprites.low)
		breadCrumbs.HPulse:Show(lowWarningPulse_glocal and isAlive and healthPerc < lowHealthThreshold_glocal)
		if shieldsMode_glocal > 0 then --this check also needs to be outside, where actual values are get
			breadCrumbs.SP:SetMax(shieldMax)
			breadCrumbs.SP:SetProgress(shieldCurr, smoothBarUpdates_glocal and shieldMax*2 or nil)
		if shieldsMode_glocal == 2 then --nested if
			breadCrumbs.SPulse:Show(lowWarningPulse_glocal and isAlive and shieldPerc < 0.3)
			breadCrumbs.Overload:Show(unit:IsShieldOverloaded())
		end end
		-- breadCrumbs.AP:SetMax(absorbMax > 0 and absorbMax or 100000)
		breadCrumbs.AP:SetMax(absorbMax)
		breadCrumbs.AP:SetProgress(absorbCurr, smoothBarUpdates_glocal and absorbMax*2 or nil)
	end
end

---------------------------------------------------------------------------------------------------
-- BreadCrumbs Options Helpers
---------------------------------------------------------------------------------------------------
local function setupSlider(path, slider, display, savedValue, prefix, suffix, formatStyle, multiplier)
	if not display then display = s_gsub(slider, "_Slider", "_Display") end
	
	local temp = savedValue; prefix = prefix or ""; suffix = suffix or ""
	if formatStyle then multiplier = multiplier or 1; temp = s_format(formatStyle, savedValue*multiplier) end
	
	slider = path:FindChild(slider); display = path:FindChild(display)
	display:SetText(prefix..temp..suffix); slider:SetValue(savedValue); slider:SetData({display, prefix, suffix, formatStyle, multiplier})
		
	return slider
end

local function setupRadioTrio(path, radioCtrl, radioOn, radioOff, radioAdv, savedValue)
	if not radioOff then radioOff = s_gsub(radioOn, "_On", "_Off") end
	
	radioOn = path:FindChild(radioOn); radioOff = path:FindChild(radioOff); radioAdv = path:FindChild(radioAdv); radioCtrl = path:FindChild(radioCtrl); radioCtrl:SetData({radioOn, radioOff}) 
	if savedValue == 1 then radioOn:SetCheck(true) elseif savedValue == 0 then radioOff:SetCheck(true) else radioAdv:SetCheck(true) end
	
	return radioCtrl
end

local function setupRadioDuo(path, radioOn, radioOff, savedValue)
	if not radioOff then radioOff = radioOn.."_Off" end
	
	radioOn = path:FindChild(radioOn); radioOff = path:FindChild(radioOff)
	if savedValue then radioOn:SetCheck(true) else radioOff:SetCheck(true) end
	
	return radioOn
end

local function setupCheckbox(path, checkBox, savedValue)
	checkBox = path:FindChild(checkBox)
	checkBox:SetCheck(savedValue)
	
	return checkBox
end

local function trinarySelection(radioCtrl)
	local data = radioCtrl:GetData()
	
	if data[1]:IsChecked() then return 1
	elseif data[2]:IsChecked() then return 0
	else return 2 end
end

function BreadCrumbs:_windowFunction_SliderChanged(windowHandler, windowControl, newValue)
	local data = windowHandler:GetData(); if not data then Print("No data present"); return end -- data = {displayControl, prefix, suffix, formatStyle}

	local temp = newValue; data[2] = data[2] or ""; data[3] = data[3] or ""
	if data[4] then data[5] = data[5] or 1; temp = s_format(data[4], newValue*data[5]) end

	data[1]:SetText(data[2]..temp..data[3])
end

---------------------------------------------------------------------------------------------------
-- BreadCrumbs Options
---------------------------------------------------------------------------------------------------
function BreadCrumbs:OnConfigure() self:_windowFunction_ProcessMenu() end

function BreadCrumbs:_windowFunction_ProcessMenu() --need testing; does it recreate stuff if options is opened and closed more than once per session? if so, add an if-check to skip stuff
	self.optionsWindow:SetAnchorOffsets(t_unpack(optionsTable_glocal.menuOffsets))

	local controlTable, optionsTable = 				controlsTable_glocal, optionsTable_glocal
	
	--General Tab 
	local firstTab = 							self.optionsWindow:FindChild("First_Tab")
	
	local generalOptions, generalControls = 			optionsTable.generalOptions, controlTable.generalControls
	generalControls.detailedPins = 					setupRadioDuo		(firstTab, "DetailedPins_RadioBtn", "NormalPins_RadioBtn", generalOptions.detailedPins)

	generalControls.distanceCoefficient = 				setupSlider		(firstTab, "DistanceCoefficient_Slider", nil, generalOptions.distanceCoefficient, "x", false, "%.1f")
	generalControls.maxDistance = 					setupSlider		(firstTab, "MaxDistance_Slider", nil, generalOptions.distanceThreshold, false, "m")
	generalControls.distanceFalloff = 				setupCheckbox		(firstTab, "DistanceFalloff_Checkbox", generalOptions.distanceFalloff)
	generalControls.falloffDistance = 				setupSlider		(firstTab, "FalloffDistance_Slider", nil, generalOptions.falloffDistance, false, "m")
	generalControls.falloffOpacity = 					setupSlider		(firstTab, "FalloffOpacity_Slider", nil, generalOptions.falloffOpacity, false, "%", "%.0f", 100)

	generalControls.scale = 						setupSlider		(firstTab, "Scale_Slider", nil, generalOptions.scale, false, "%", "%.0f", 100)
	generalControls.opacity = 						setupSlider		(firstTab, "Opacity_Slider", nil, generalOptions.opacity, false, "%", "%.0f", 100)
	
	generalControls.lowHealthThreshold = 			setupSlider		(firstTab, "LowHealthThreshold_Slider", "HealthBarThresholds_Display", generalOptions.lowHealthThreshold, false, "%", "%.0f", 100)
	generalControls.averageHealthThreshold = 			setupSlider		(firstTab, "AverageHealthThreshold_Slider", "HealthBarThresholds_Display", generalOptions.averageHealthThreshold, false, "%", "%.0f", 100)
	generalControls.relationshipIcons =				setupRadioDuo		(firstTab, "RelationshipIcons_RadioBtn", nil, generalOptions.relationshipIcons)
	generalControls.colorfulClassIcons = 				setupRadioDuo		(firstTab, "ColorfulClassIcons_RadioBtn", nil, generalOptions.colorfulClassIcons)
	generalControls.clickableFrames = 				setupRadioDuo		(firstTab, "ClickableFrames_RadioBtn", nil, generalOptions.clickableFrames)
	
	generalControls.shieldsMode = 					setupRadioTrio 		(firstTab, "Shields_RadioControl", "Shields_RadioBtn_On", nil, "Shields_RadioBtn_Extended", generalOptions.shieldsMode)
	generalControls.hybridMode = 					setupRadioTrio 		(firstTab, "HybridMode_RadioControl", "HybridMode_RadioBtn_On", nil, "HybridMode_RadioBtn_Critical", generalOptions.hybridMode)

	generalControls.drawOnNextFrame = 				setupRadioDuo		(firstTab, "SyncedTimer_RadioBtn", "CustomTimer_RadioBtn", generalOptions.drawOnNextFrame)
	generalControls.primaryTimer = 				setupSlider		(firstTab, "Timers_PrimaryTimer_Slider", nil, generalOptions.primaryTimer, false, "Hz")
	generalControls.secondaryTimer = 				setupSlider		(firstTab, "Timers_SecondaryTimer_Slider", nil, generalOptions.secondaryTimer, false, "Hz")

	--Tracking Options
	local secondTab = 		 					self.optionsWindow:FindChild("Second_Tab")

	local hostileOptions, hostileControls = 			optionsTable.breadOptions.hostile, controlTable.breadControls.hostile
	hostileControls.enable = 						setupCheckbox		(secondTab, "Hostile_Enable_Checkbox", hostileOptions.enabled)
	hostileControls.ignoreOffscreen = 				setupCheckbox		(secondTab, "Hostile_IgnoreOffscreen_Checkbox", hostileOptions.ignoreOffscreen)
	hostileControls.ignoreOccluded = 				setupCheckbox		(secondTab, "Hostile_IgnoreOccluded_Checkbox", hostileOptions.ignoreOccluded)
	hostileControls.ignoreDead = 					setupCheckbox		(secondTab, "Hostile_IgnoreDead_Checkbox", hostileOptions.ignoreDead)
	hostileControls.maximumAmount = 				setupSlider		(secondTab, "Hostile_MaximumAmount_Slider", nil, hostileOptions.maximumAmount)
	hostileControls.ignoreAbove = 					setupCheckbox		(secondTab, "Hostile_IgnoreAbove_Checkbox"); hostileControls.ignoreAbove:SetCheck(hostileOptions.ignoreAbove)
	hostileControls.ignoreThreshold = 				setupSlider		(secondTab, "Hostile_IgnoreThreshold_Slider", nil, hostileOptions.ignoreThreshold, false, "%", "%.0f", 100)

	local allyOptions, allyControls = 				optionsTable.breadOptions.ally, controlTable.breadControls.ally
	allyControls.enable = 						setupCheckbox		(secondTab, "Ally_Enable_Checkbox", allyOptions.enabled)
	allyControls.ignoreOffscreen = 					setupCheckbox		(secondTab, "Ally_IgnoreOffscreen_Checkbox", allyOptions.ignoreOffscreen)
	allyControls.ignoreOccluded = 					setupCheckbox		(secondTab, "Ally_IgnoreOccluded_Checkbox", allyOptions.ignoreOccluded)
	allyControls.ignoreDead = 						setupCheckbox		(secondTab, "Ally_IgnoreDead_Checkbox", allyOptions.ignoreDead)
	allyControls.maximumAmount = 				setupSlider		(secondTab, "Ally_MaximumAmount_Slider", nil, allyOptions.maximumAmount)
	allyControls.ignoreAbove = 					setupCheckbox		(secondTab, "Ally_IgnoreAbove_Checkbox"); allyControls.ignoreAbove:SetCheck(allyOptions.ignoreAbove)
	allyControls.ignoreThreshold = 					setupSlider		(secondTab, "Ally_IgnoreThreshold_Slider", nil, allyOptions.ignoreThreshold, false, "%", "%.0f", 100)
	
	local groupOptions, groupControls = 				optionsTable.breadOptions.group, controlTable.breadControls.group
	groupControls.enable = 						setupCheckbox		(secondTab, "Group_Enable_Checkbox", groupOptions.enabled)
	groupControls.ignoreOffscreen = 				setupCheckbox		(secondTab, "Group_IgnoreOffscreen_Checkbox", groupOptions.ignoreOffscreen)
	groupControls.ignoreOccluded = 				setupCheckbox		(secondTab, "Group_IgnoreOccluded_Checkbox", groupOptions.ignoreOccluded)
	groupControls.ignoreDead = 					setupCheckbox		(secondTab, "Group_IgnoreDead_Checkbox", groupOptions.ignoreDead)
	groupControls.maximumAmount = 				setupSlider		(secondTab, "Group_MaximumAmount_Slider", nil, groupOptions.maximumAmount)
	groupControls.ignoreAbove = 					setupCheckbox		(secondTab, "Group_IgnoreAbove_Checkbox"); groupControls.ignoreAbove:SetCheck(groupOptions.ignoreAbove)
	groupControls.ignoreThreshold = 				setupSlider		(secondTab, "Group_IgnoreThreshold_Slider", nil, groupOptions.ignoreThreshold, false, "%", "%.0f", 100)
	
	local focusOptions, focusControls = 				optionsTable.breadOptions.focus, controlTable.breadControls.focus
	focusControls.enable = 						setupCheckbox		(secondTab, "Focus_Enable_Checkbox", focusOptions.enabled)
	focusControls.ignoreOffscreen = 				setupCheckbox		(secondTab, "Focus_IgnoreOffscreen_Checkbox", focusOptions.ignoreOffscreen)
	focusControls.ignoreOccluded = 					setupCheckbox		(secondTab, "Focus_IgnoreOccluded_Checkbox", focusOptions.ignoreOccluded)
	focusControls.ignoreDead = 					setupCheckbox		(secondTab, "Focus_IgnoreDead_Checkbox", focusOptions.ignoreDead)
	focusControls.ignoreAbove = 					setupCheckbox		(secondTab, "Focus_IgnoreAbove_Checkbox"); focusControls.ignoreAbove:SetCheck(focusOptions.ignoreAbove)
	focusControls.ignoreThreshold = 				setupSlider		(secondTab, "Focus_IgnoreThreshold_Slider", nil, focusOptions.ignoreThreshold, false, "%", "%.0f", 100)
	
	generalControls.crumb = 						setupCheckbox		(secondTab, "Crumbs_Enable_Checkbox", generalOptions.crumbs)
	generalControls.crumbInstantTransition = 			setupRadioDuo		(secondTab, "InstantTransition_RadioBtn", "SmoothTransition_RadioBtn", generalOptions.crumbInstantTransition)
	generalControls.crumbAmount = 				setupSlider		(secondTab, "Crumbs_MaximumAmount_Slider", nil, generalOptions.crumbAmount)
	generalControls.crumbDistance = 				setupSlider		(secondTab, "Crumbs_DistanceInterval_Slider", nil, generalOptions.crumbTick, false, "m")

	firstTab:AttachTab(secondTab, false)
	firstTab:Lock(true); secondTab:Lock(true)
	self.optionsWindow:Invoke()
end

function BreadCrumbs:_windowFunction_OK()
	local controlTable, optionsTable = 				controlsTable_glocal, optionsTable_glocal

	local generalControls, generalOptions = 			controlTable.generalControls, optionsTable.generalOptions
	generalOptions.detailedPins = 					generalControls.detailedPins:IsChecked()

	generalOptions.distanceCoefficient =				generalControls.distanceCoefficient:GetValue()
	generalOptions.distanceThreshold =				generalControls.maxDistance:GetValue()
	generalOptions.distanceFalloff = 					generalControls.distanceFalloff:IsChecked()
	generalOptions.falloffDistance = 					generalControls.falloffDistance:GetValue()
	generalOptions.falloffOpacity = 					generalControls.falloffOpacity:GetValue()

	generalOptions.scale =						generalControls.scale:GetValue()
	generalOptions.opacity =						generalControls.opacity:GetValue()

	generalOptions.lowHealthThreshold = 				generalControls.lowHealthThreshold:GetValue()
	generalOptions.averageHealthThreshold = 			generalControls.averageHealthThreshold:GetValue()
	generalOptions.relationshipIcons = 				generalControls.relationshipIcons:IsChecked()
	generalOptions.colorfulClassIcons = 				generalControls.colorfulClassIcons:IsChecked()
	generalOptions.clickableFrames = 				generalControls.clickableFrames:IsChecked()
	
	generalOptions.shieldsMode = 					trinarySelection(generalControls.shieldsMode)
	generalOptions.hybridMode = 					trinarySelection(generalControls.hybridMode)

	generalOptions.drawOnNextFrame = 				generalControls.drawOnNextFrame:IsChecked()
	generalOptions.primaryTimer =					generalControls.primaryTimer:GetValue()
	generalOptions.secondaryTimer =				generalControls.secondaryTimer:GetValue()

	local hostileControls, hostileOptions = 			controlTable.breadControls.hostile, optionsTable.breadOptions.hostile
	hostileOptions.enabled =						hostileControls.enable:IsChecked()
	hostileOptions.ignoreOffscreen = 				hostileControls.ignoreOffscreen:IsChecked()
	hostileOptions.ignoreOccluded = 					hostileControls.ignoreOccluded:IsChecked()
	hostileOptions.ignoreDead = 					hostileControls.ignoreDead:IsChecked()
	hostileOptions.maximumAmount =				hostileControls.maximumAmount:GetValue()
	hostileOptions.ignoreAbove =					hostileControls.ignoreAbove:IsChecked()
	hostileOptions.ignoreThreshold =					hostileControls.ignoreThreshold:GetValue()

	local allyControls, allyOptions = 				controlTable.breadControls.ally, optionsTable.breadOptions.ally
	allyOptions.enabled =							allyControls.enable:IsChecked()
	allyOptions.ignoreOffscreen = 					allyControls.ignoreOffscreen:IsChecked()
	allyOptions.ignoreOccluded = 					allyControls.ignoreOccluded:IsChecked()
	allyOptions.ignoreDead = 						allyControls.ignoreDead:IsChecked()
	allyOptions.maximumAmount =					allyControls.maximumAmount:GetValue()
	allyOptions.ignoreAbove =						allyControls.ignoreAbove:IsChecked()
	allyOptions.ignoreThreshold =					allyControls.ignoreThreshold:GetValue()
	
	local groupControls, groupOptions = 				controlTable.breadControls.group, optionsTable.breadOptions.group
	groupOptions.enabled =						groupControls.enable:IsChecked()
	groupOptions.ignoreOffscreen = 					groupControls.ignoreOffscreen:IsChecked()
	groupOptions.ignoreOccluded = 					groupControls.ignoreOccluded:IsChecked()
	groupOptions.ignoreDead = 					groupControls.ignoreDead:IsChecked()
	groupOptions.maximumAmount =				groupControls.maximumAmount:GetValue()
	groupOptions.ignoreAbove =					groupControls.ignoreAbove:IsChecked()
	groupOptions.ignoreThreshold =					groupControls.ignoreThreshold:GetValue()

	local focusControls, focusOptions = 				controlTable.breadControls.focus, optionsTable.breadOptions.focus
	focusOptions.enabled =						focusControls.enable:IsChecked()
	focusOptions.ignoreOffscreen = 					focusControls.ignoreOffscreen:IsChecked()
	focusOptions.ignoreOccluded = 					focusControls.ignoreOccluded:IsChecked()
	focusOptions.ignoreDead = 					focusControls.ignoreDead:IsChecked()
	focusOptions.ignoreAbove =					focusControls.ignoreAbove:IsChecked()
	focusOptions.ignoreThreshold =					focusControls.ignoreThreshold:GetValue()
	
	generalOptions.crumbs =						generalControls.crumb:IsChecked()
	generalOptions.crumbInstantTransition = 			generalControls.crumbInstantTransition:IsChecked()
	generalOptions.crumbAmount =					generalControls.crumbAmount:GetValue()
	generalOptions.crumbTick =					generalControls.crumbDistance:GetValue()
	
	optionsTable.menuOffsets = {self.optionsWindow:GetAnchorOffsets()}

	self.optionsWindow:Close()
	RequestReloadUI()
end

function BreadCrumbs:_windowFunction_Cancel() self.optionsWindow:Close() end

function BreadCrumbs:_windowFunction_RestoreDefaults() --cancel hold signal
	optionsTable_glocal.restoreDefaults = true

	self.optionsWindow:Close()
	RequestReloadUI()
end

function BreadCrumbs:_windowFunction_BreadClicked(windowHandler, _, mouseButton)
	local clickedUnit = windowHandler:GetParent():GetData()
	local clickedUnitId = clickedUnit:GetId()
	local distance = unitsTable_glocal[unitsMasterTable_glocal[clickedUnitId]][clickedUnitId].distance
	
	if distance < 100 then
		if mouseButton == 0 then GameLib.SetTargetUnit(clickedUnit)
		elseif mouseButton == 1 then GameLib.SetTargetUnit(clickedUnit:GetTarget()) end
	end
	--if Apollo.IsAltKeyDown() then Print("alt") elseif Apollo.IsControlKeyDown() then Print("ctrl") elseif Apollo.IsShiftKeyDown() then Print("shift") else ---- end
end

-----------------------------------------------------------------------------------------------
-- BreadCrumbs Instance
-----------------------------------------------------------------------------------------------
local BreadCrumbsInst = BreadCrumbs:new()
BreadCrumbsInst:Init()